/*
 *	Sherlock Gatherer -- Databases
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <gdbm.h>

#include "gather.h"

/** CONFIGURATION **/

static uns urldb_cache_size, md5db_cache_size;

static struct cfitem cfitems[] = {
  { "URLDbCacheSize", CI_INT, &urldb_cache_size },
  { "MD5DbCacheSize", CI_INT, &md5db_cache_size },
  { NULL, CI_STOP, NULL }
};

void
db_init(void)
{
  cf_read("cf/dbm", cfitems);
}

/** THE GATHERER LOCK ***/

void
gather_lock(void)
{
  int fd;

  fd = open(LOCK_NAME, O_WRONLY | O_CREAT | O_EXCL, 0666);
  if (fd < 0)
    die("Cannot get gatherer lock: %m");
}

void
gather_unlock(void)
{
  if (unlink(LOCK_NAME) < 0)
    die("Cannot free gatherer lock: %m");
}

/*** THE URL DATABASE ***/

static GDBM_FILE urldb;

void
urldb_open(void)
{
  if (! (urldb = gdbm_open(URLDB_NAME, 0, GDBM_WRCREAT + GDBM_FAST, 0666, NULL)))
    die("Cannot open the URL database, err=%d", gdbm_errno);
  if (urldb_cache_size && gdbm_setopt(urldb, GDBM_CACHESIZE, &urldb_cache_size, sizeof(urldb_cache_size)))
    die("Cannot set URL database cache size");
}

void
urldb_close(void)
{
  gdbm_close(urldb);
}

struct urlrec *
urldb_lookup(byte *url)
{
  datum KEY = { (void *) url, strlen(url) };
  datum UR;
  static struct urlrec ur;

  UR = gdbm_fetch(urldb, KEY);
  if (UR.dptr)
    {
      memcpy(&ur, UR.dptr, sizeof(ur));
      free(UR.dptr);
      return &ur;
    }
  else
    return NULL;
}

int
urldb_exists(byte *url)
{
  datum KEY = { (void *) url, strlen(url) };

  return gdbm_exists(urldb, KEY);
}

static datum urldb_gkey;

byte *
urldb_get_first(byte *buf)
{
  if (urldb_gkey.dptr)
    free(urldb_gkey.dptr);
  urldb_gkey = gdbm_firstkey(urldb);
  if (!urldb_gkey.dptr)
    return NULL;
  memcpy(buf, urldb_gkey.dptr, urldb_gkey.dsize);
  buf[urldb_gkey.dsize] = 0;
  return buf;
}

byte *
urldb_get_next(byte *buf)
{
  void *z = urldb_gkey.dptr;
  urldb_gkey = gdbm_nextkey(urldb, urldb_gkey);
  if (z)
    free(z);
  if (!urldb_gkey.dptr)
    return NULL;
  memcpy(buf, urldb_gkey.dptr, urldb_gkey.dsize);
  buf[urldb_gkey.dsize] = 0;
  return buf;
}

int
urldb_delete(byte *url)
{
  datum KEY = { (void *) url, strlen(url) };

  return !gdbm_delete(urldb, KEY);
}

void
urldb_store(byte *url, ulg stat, ulg oid)
{
  struct urlrec ur = { stat, oid };
  datum UR = { (void *) &ur, sizeof(ur) };
  datum KEY = { (void *) url, strlen(url) };
  int err;

  if (err = gdbm_store(urldb, KEY, UR, GDBM_REPLACE))
    die("urldb_store: gdbm_store returned %d", err);
}

/*** THE MD5 SUM DATABASE ***/

static GDBM_FILE md5db;

void
md5db_open(void)
{
  if (! (md5db = gdbm_open(MD5DB_NAME, 0, GDBM_WRCREAT + GDBM_FAST, 0666, NULL)))
    die("Cannot open the MD5 database, err=%d", gdbm_errno);
  if (md5db_cache_size && gdbm_setopt(md5db, GDBM_CACHESIZE, &md5db_cache_size, sizeof(md5db_cache_size)))
    die("Cannot set MD5 database cache size");
}

void
md5db_close(void)
{
  gdbm_close(md5db);
}

struct md5rec *
md5db_lookup(byte *md5)
{
  datum KEY = { (void *) md5, MD5_SIZE };
  datum MD;
  static struct md5rec md;

  MD = gdbm_fetch(md5db, KEY);
  if (MD.dptr)
    {
      memcpy(&md, MD.dptr, sizeof(md));
      free(MD.dptr);
      return &md;
    }
  else
    return NULL;
}

int
md5db_exists(byte *md5)
{
  datum KEY = { (void *) md5, MD5_SIZE };

  return gdbm_exists(md5db, KEY);
}

static datum md5db_gkey;

byte *
md5db_get_first(void)
{
  if (md5db_gkey.dptr)
    free(md5db_gkey.dptr);
  md5db_gkey = gdbm_firstkey(md5db);
  if (!md5db_gkey.dptr)
    return NULL;
  else if (md5db_gkey.dsize != MD5_SIZE)
    die("md5db_get_first: invalid size of key");
  return md5db_gkey.dptr;
}

byte *
md5db_get_next(void)
{
  void *z = md5db_gkey.dptr;
  md5db_gkey = gdbm_nextkey(md5db, md5db_gkey);
  if (z)
    free(z);
  if (!md5db_gkey.dptr)
    return NULL;
  else if (md5db_gkey.dsize != MD5_SIZE)
    die("md5db_get_first: invalid size of key");
  return md5db_gkey.dptr;
}

int
md5db_delete(byte *md5)
{
  datum KEY = { (void *) md5, MD5_SIZE };

  return !gdbm_delete(md5db, KEY);
}

void
md5db_store(byte *md5, ulg oid)
{
  struct md5rec md = { oid };
  datum MD = { (void *) &md, sizeof(md) };
  datum KEY = { (void *) md5, MD5_SIZE };
  int err;

  if (err = gdbm_store(md5db, KEY, MD, GDBM_REPLACE))
    die("md5db_store: gdbm_store returned %d", err);
}
