/*
 *	Sherlock Gatherer -- Main Program
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <time.h>
#include <signal.h>
#include <errno.h>

#include "gather.h"

#include <lib/url.h>
#include <lib/string.h>

/* Config */

uns max_threads = 8;
uns max_queued_objs = 16;
uns trace_jobs = 0;
uns min_server_delay = 5;
uns robot_file_timeout = 86400 * 7;
uns rec_err_dly1 = 60;
uns rec_err_limit = 3;
uns rec_err_dly2 = 3600;
uns downgrade_timeout = 1000000;
uns clock_drift = 3600;
uns trace_refs = 0;
uns refresh_after = 0;
uns dont_unlink = 0;
byte *downloaders[URL_PROTO_MAX];
uns max_queue_time = ~0;
uns max_rec_err = ~0;
uns auto_enqueue_root = 0;
uns show_forbidden = 1;

/* Jobs */

struct job {
  node n;
  pid_t pid;
  struct qhost *host;
  struct odes *hobj;
  struct qobj *qobj;
  struct odes *obj;
  int stat;
  int fetching_robot_file;
  ulg refetching;			/* Object ID we try to refetch */
  byte url[MAX_URL_SIZE];
  byte md5[MD5_SIZE];
};

#define JS_FREE 0
#define JS_GATHER 1
#define JS_ANALYSE 2

static struct job fg_job_struct;
static struct job *fg_job;
static list free_jobs, running_jobs;

static struct tempfile temp_obj;

static volatile uns shut_down;

static void
job_init(void)
{
  struct job *j;
  uns i;

  init_list(&free_jobs);
  init_list(&running_jobs);
  for(i=0; i<max_threads; i++)
    {
      j = xmalloc(sizeof(struct job));
      j->stat = JS_FREE;
      add_tail(&free_jobs, &j->n);
    }
}

void					/* Unlink or move to "old" */
maybe_unlink(byte *name)
{
  byte nn[256];
  byte *z;

  if (!dont_unlink)
    {
      unlink(name);
      return;
    }

  z = strrchr(name, '/');
  strcpy(nn, "old/");
  strcat(nn, (z ? z+1 : name));
  if (rename(name, nn))
    {
      log(L_INFO "Deleting %s", nn);
      unlink(nn);
      if (rename(name, nn))
	log(L_ERROR "Cannot rename %s to %s", name, nn);
    }
}

static void
discard_objects(struct job *j)
{
  struct oattr *a;

  for(a = find_attr(j->obj, '@'); a; a=a->same)	/* Delete data file if there's any */
    maybe_unlink(a->val);

  obj_free(j->obj);			/* Free cached object data */
  qobj_del(j->qobj);			/* Delete the temp file */
}

static void
end_fg_job(uns oid)
{
  struct job *j = fg_job;
  byte *url;
  time_t tim = time(NULL);

  if (! (url = find_aval(j->obj, 'U')))
    log(L_ERROR "No URL in end_fg_job");
  else
    {
      if (trace_jobs)
	log(L_DEBUG "%s: done (%08x)", j->qobj->tf.name, oid);
      urldb_store(url, tim & ~1, oid);
    }

  if (temp_obj.name[0])
    {
      if (temp_obj.fh)
	close(temp_obj.fh);
      maybe_unlink(temp_obj.name);
      temp_obj.fh = 0;
      temp_obj.name[0] = 0;
    }
  discard_objects(j);
  fg_job = NULL;
}

static void
do_ref(byte *r, int force)
{
  byte buf1[MAX_URL_SIZE], buf2[MAX_URL_SIZE], host[MAX_URL_SIZE], url[MAX_URL_SIZE], urlx[MAX_URL_SIZE];
  struct url u;
  uns err, rfsh, tim;
  struct qhost *h;
  struct urlrec *ur;
  int add_root = 0;

  if (trace_refs)
    log(L_DEBUG "Ref: %s", r);

  if (err = url_canon_split(r, buf1, buf2, &u))
    {
      log(L_ERROR "REF %s: %s", r, url_error(err));
      return;
    }

  if (!u.protoid || u.protoid >= URL_PROTO_MAX || !downloaders[u.protoid])
    {
      if (show_forbidden)
	log(L_INFO "%s dropped: unknown protocol", r);
      return;
    }

  tim = time(NULL);

restart:
  if ((err = url_pack(&u, urlx)) || (err = url_enescape(urlx, url)))
    {
      log(L_ERROR "%s: %s", r, url_error(err));
      return;
    }
  if (translate_url(&u, url, buf1, buf2) < 0)
    return;

  if (h = find_host(u.protoid, u.host, u.port))	/* Host known by this name -> OK */
    ;
  else if (normalize_host_name(u.host, host)) /* Host name requires normalization */
    {
      u.host = host;
      /* XXX: Mark it as an alias */
      h = find_host(u.protoid, u.host, u.port);
    }

  if (! okay_to_get(NULL, &u))
    {
      if (show_forbidden)
	log(L_INFO "%s: forbidden", url);
      return;
    }

  if (ur = urldb_lookup(url))		/* Consult the URL database */
    {
      if (ur->access & 1)
	{
	  if (trace_refs)
	    log(L_DEBUG "%s: already queued", url);
	  return;
	}
      rfsh = force ? 60 : refresh_after;
      if (rfsh && ur->access + rfsh <= tim)
	{
	  if (trace_refs)
	    log(L_DEBUG "%s: known, but refreshing", url);
	}
      else
	{
	  if (trace_refs)
	    log(L_DEBUG "%s: already known", url);
	  return;
	}
    }
  urldb_store(url, tim | 1, ur ? ur->oid : 0);

  if (!h)				/* New host? */
    {
      struct odes *o = obj_new();
      h = new_host(u.protoid, u.host, u.port, new_host_id());
      set_attr(o, 'H', u.host);
      set_attr_num(o, 'P', u.port);
      set_attr_num(o, 'p', u.protoid);
      dump_host_obj(h->host_id, o);
      obj_free(o);
      if (auto_enqueue_root && strcmp(u.rest, "/"))
	add_root = 1;
    }

  enqueue_item(h, u.rest);		/* Enqueue it and kick it to the queue */
  enqueue_host(h);

  if (add_root)				/* Add also server root if requested to do so */
    {
      strcpy(u.rest, "/");
      add_root = 0;
      if (trace_refs)
	log(L_DEBUG "Adding ref to root");
      goto restart;
    }
}

static void
do_file_ref(byte *r)
{
  if (trace_refs)
    log(L_DEBUG "File ref to %s", r);
  qobj_enqueue_name(r);
}

static void				/* Extract references and finish processing */
do_refs(uns oid)
{
  struct job *j = fg_job;
  struct odes *d = j->obj;
  struct oattr *a;

  for(a=find_attr(d, 'R'); a; a=a->same) /* URL refs */
    do_ref(a->val, 0);

  for(a=find_attr(d, 'F'); a; a=a->same) /* File refs */
    do_file_ref(a->val);

  end_fg_job(oid);
}

static void
subst_and_run(struct odes *d, byte *c)	/* Substitute parameters and exec */
{
  char *args[32];
  byte buf[1024];
  uns ac = 0;
  byte *k = buf;
  byte *e = buf + sizeof(buf) - 5;
  byte *oc = c;

  for(;;)
    {
      while (Cspace(*c))
	c++;
      if (!*c)
	break;
      args[ac++] = k;
      if (ac > 30)
	die("`%s': Too many args", oc);
      while (*c && !Cspace(*c))
	{
	  if (k > e)
	    {
	    over:
	      die("`%s': Arg overflow", oc);
	    }
	  if (*c == '%')
	    {
	      if (c[1] == '%')
		*k++ = *c++, c++;
	      else
		{
		  uns x, y;
		  byte *z;
		  if (*++c == '?' && c[0] && c[1])
		    c++, x = *c++, y=*c++;
		  else
		    x = *c++, y = 0;
		  if (!x)
		    die("`%s': Invalid substitution", oc);
		  z = find_aval(d, x);
		  if (!z && y)
		    z = find_aval(d, y);
		  if (!z)
		    die("`%s': Attribute `%c' not found", oc, y ? y : x);
		  x = strlen(z);
		  if (x + k > e)
		    goto over;
		  strcpy(k, z);
		  k += x;
		}
	    }
	  else
	    *k++ = *c++;
	}
      *k++ = 0;
    }
  if (!ac)
    die("`%s': Missing program name", oc);
  args[ac] = NULL;
  if (trace_jobs)
    {
      byte z[1024];
      uns i;
      k = z;
      for(i=0; i<ac; i++)
	{
	  if (i)
	    *k++ = ' ';
	  strcpy(k, args[i]);
	  k += strlen(args[i]);
	}
      *k = 0;
      log(L_DEBUG "Running %s", z);
    }
  execv(args[0], args);
}

static void				/* Add another URL to an object */
add_obj_url(ulg oid, byte *url)
{
  byte name[64];
  FILE *f;

  mk_obj_name("obj", oid, name);	/* We circumvent the standard obj manipulation */
  f = fopen(name, "a");			/* routines to gain some speed :-( */
  fprintf(f, "U%s\n", url);
  fclose(f);
}

static int				/* Start processing of queued object */
run_fg_job(void)
{
  struct job *j;
  struct qobj *o;
  struct odes *d;
  byte *b, *m;
  struct oattr *a;
  int err, pid, fd2;
  struct md5rec *mdr;
  FILE *f;

#if 0
  queue_sync();
#endif

  if (! (o = qobj_dequeue()))		/* Any work for us? */
    return 0;

  j = fg_job = &fg_job_struct;		/* Run the foreground job... */
  j->pid = 0;				/* No special processes yet */
  j->host = NULL;			/* We work with no explicit host ... */
  j->hobj = NULL;			/* ... and have no buffered host objects */
  j->qobj = o;
  if (!o->obj)				/* No object info cached => load it */
    o->obj = obj_load(o->tf.name);
  d = j->obj = o->obj;			/* The cached object we're working with */
  j->stat = JS_ANALYSE;

  for (a = find_attr(d, '.'); a; a=a->same) /* Extract comments */
    log(L_INFO "%s: %s", o->tf.name, a->val);

  if (b = find_aval(d, 'X'))		/* Object has no contents */
    {
      log(L_INFO "%s: no contents", o->tf.name);
      do_refs(8888);
      return 1;
    }

  while (err = datatype_guess(d))	/* Process data types... */
    {
      if (err < 0)
	err = c_e_decode(d, err);	/* Content-decoding? */
      if (err)
	{
	  do_refs(err);
	  return 1;
	}
    }

  if (! (b = find_aval(d, 'T')))	/* Decide what to do with it... */
    die("datatype_guess forgot to report a data type attribute");
  if (! (b = lookup_distiller(b)))
    {
      log(L_INFO "%s: Don't know how to handle", find_aval(d, 'U'));
      do_refs(302);
      return 1;
    }

  if (! (m = find_aval(d, 'M')))	/* Consult the MD5 database */
    die("datatype_guess returned no MD5 sum");
  hex_to_md5(m, j->md5);
  if (mdr = md5db_lookup(j->md5))	/* Already exists... */
    {
      byte *url = find_aval(d, 'U');
      ulg oid = mdr->oid;
      log(L_INFO "%s: Already known as object %08x", url, oid);
      add_obj_url(oid, url);
      do_refs(oid);
      return 1;
    }

  open_temp(&temp_obj, TF_OBJECT);	/* Dump object data to the temp file */
  if ((fd2 = dup(temp_obj.fh)) < 0)
    die("dup: %m");
  if (! (f = fdopen(fd2, "w")))
    die("fdopen: %m");
  if (trace_jobs > 1)
    log(L_DEBUG "Dumping temporary data to %s", temp_obj.name);
  obj_fwrite(f, d);
  fclose(f);

  pid = fork();				/* And start the distiller process */
  if (pid < 0)
    die("fork() failed: %m");
  if (pid > 0)
    j->pid = pid;
  else
    {
      byte *from = find_aval(d, '@');
      initlog(NULL);			/* Re-read PID */
      close(0);
      if (!from)
	die("Distiller process: no infile");
      if (open(from, O_RDONLY, 0) != 0)
	die("Distiller: open(from): %m");
      close(1);
      if (dup(temp_obj.fh) != 1)
	die("Distiller: dup(dest): %m");
      lseek(1, 0, SEEK_END);
      subst_and_run(d, b);
      die("exec [%s]: %m", b);
    }
  return 1;
}

static void
clean_up_obj(struct odes *d)
{
	/* We cannot do this here as it breaks cleanup routines! */
#if 0
  set_attr(d, '@', NULL);		/* Kill all temporary attributes before dumping the object */
#endif
}

static void
finished_fg_job(uns stat)
{
  struct job *j = fg_job;
  FILE *f;
  byte *b, *url;
  struct odes *d;
  ulg oid;

  if (trace_jobs)
    log(L_DEBUG "FG job %d exited with status %d", j->pid, stat);
  j->pid = 0;
  obj_free(j->obj);			/* Re-load the object */
  lseek(temp_obj.fh, 0, SEEK_SET);
  if (! (f = fdopen(temp_obj.fh, "r")))
    die("fdopen: %m");
  j->obj = d = obj_fload(f, temp_obj.name);
  fclose(f);
  temp_obj.fh = 0;

  if (! (url = find_aval(d, 'U')))	/* Find URL */
    die("finished_bg_job: No URL in object descriptor");

  if (b = find_aval(d, '-'))		/* Chewing error */
    sscanf(b, "%d", &oid);
  else if (b = find_aval(d, 'X'))	/* No contents */
    oid = 8888;
  else					/* Process contents */
    {
      uns retry = 0;
      clean_up_obj(d);
      for(;;)				/* Find unused object ID */
	{
	  oid = new_oid(retry++);
	  if (dump_obj_to_file("obj", oid, d, 1))
	    break;
	  if (retry++ > 10000)
	    die("finished_fg_job: max_retry exceeded");
	}
    }

  md5db_store(j->md5, oid);		/* Update MD5 database */
  do_refs(oid);				/* Process all references and proceed with next object */
}

static void
delete_object(ulg oid, byte *url)
{
  byte name[64];
  struct odes *d;
  struct oattr *a;
  byte *x;

  if (oid < OID_MIN)
    return;
  mk_obj_name("obj", oid, name);
  d = obj_load(name);
  a = find_attr(d, 'U');
  while (a)
    {
      if (!strcmp(a->val, url))
	break;
      a = a->same;
    }
  if (!a)
    {
      log(L_ERROR "%s: object %08x doesn't contain this URL", url, oid);
      goto erex;
    }
  if (!del_attr(d, a))
    {
      log(L_ERROR "%s: del_attr failed", url);
      goto erex;
    }
  a = find_attr(d, 'U');
  if (a)				/* More references -> enqueue them all, they're probably invalid */
    {
      struct oattr *b;
      int cnt = 0;
      obj_write(name, d);
      for(b=a; b; b=b->same)
	cnt++;
      if (cnt < 5)			/* Magic limit */
	while (a)
	  {
	    do_ref(a->val, 1);
	    a = a->same;
	  }
    }
  else					/* No more refs => delete the object file */
    {
      if (x = find_aval(d, 'M'))
	{
	  byte md5[MD5_SIZE];
	  hex_to_md5(x, md5);
	  if (!md5db_delete(md5))
	    {
	      log(L_ERROR "Error while deleting MD5 %s for %s", x, url);
	      goto erex;
	    }
	}
      if (unlink(name) < 0)
	log(L_ERROR "unlink(%s) failed: %m", name);
      if (trace_jobs)
	log(L_DEBUG "Object %08x replaced", oid);
    }

erex:
  obj_free(d);
}

static void
run_bg_job(struct qhost *h)
{
  struct job *j = HEAD(free_jobs);
  pid_t pid;
  struct url u;
  struct qitem *i;
  uns err, dly, del, rob, laccess;
  byte *down, *b;
  byte buf[MAX_URL_SIZE];
  struct odes *d;
  uns if_modified_since = 0;
  uns downgrade = 0;
  struct urlrec *urr;
  uns now = time(NULL);

  if (trace_jobs)
    log(L_DEBUG "Starting job for %s:%d:%d", h->name, h->protocol, h->port);

  j->host = h;				/* Load host description file */
  d = j->hobj = load_host_obj(h->host_id);

  if (b = find_aval(d, 'L'))		/* Check last access time */
    {
      uns del;				/* Too fast? */
      laccess = atol(b);
      del = (b = find_aval(d, 'D')) ? atol(b) : min_server_delay;
      dly = laccess + del;
    }
  else
    {
      laccess = now - 1;
      dly = 0;
    }

  if (b = find_aval(d, 'O'))		/* Out of service until */
    {
      del = atol(b);
      if (del > dly)
	dly = del;
      else
	set_attr(d, 'O', NULL);
    }

  if (b = find_aval(d, 'C'))		/* Last access was a recoverable error? */
    {
      del = laccess + (((uns) atol(b) >= rec_err_limit) ? rec_err_dly2 : rec_err_dly1);
      if (del > dly)
	dly = del;
    }

  if (b = find_aval(d, 'd'))		/* Downgrade protocol? */
    {
      uns dpt = atol(b);
      if (dpt + downgrade_timeout > now)
	downgrade = 1;
      else
	set_attr(d, 'd', NULL);
    }

  if (dly > now)			/* Unfortunately, we must wait... */
    {
      if (trace_jobs)
	log(L_DEBUG "Delayed for %d seconds", dly - now);
      obj_free(d);
      wait_host(h, dly);
      return;
    }

restart0:
  j->refetching = 0;
  rob = (b = find_aval(d, 'r')) ? atol(b) : robot_file_timeout; /* What about the `robots.txt' file? */
  if (rob && h->protocol == URL_PROTO_HTTP)
    {
      if (b = find_aval(d, 'R'))
	{
	  uns rl = atol(b);
	  j->fetching_robot_file = rl + rob < now;
	}
      else
	j->fetching_robot_file = 1;
    }
  else
    j->fetching_robot_file = 0;

 restart:				/* Get one item off the queue */
  if (j->fetching_robot_file)
    u.rest = "/robots.txt";
  else
    {
      if (! (i = peek_item(h)) || shut_down)
	{
	  if (!i && trace_jobs)
	    log(L_DEBUG "No more items for this server");
	  obj_free(d);
	  enqueue_host(h);
	  return;
	}
      u.rest = i->text;
    }

  u.protoid = h->protocol;		/* Build the URL and check it */
  if (u.protoid >= URL_PROTO_MAX)
    u.protoid = 0;
  u.protocol = url_proto_names[u.protoid];
  u.user = NULL;
  u.host = h->name;
  u.port = h->port;
  if ((err = url_pack(&u, buf)) || (err = url_enescape(buf, j->url)))
    die("run_bg_job: URL processing error: %s", url_error(err));
  if (trace_jobs)
    log(L_DEBUG "URL = %s", j->url);

  urr = urldb_lookup(j->url);		/* Do we know that object? */
  if (urr)
    {
      if (urr->oid >= OID_MIN)
	{
	  if (trace_jobs > 3)
	    log(L_DEBUG "Refetch for OID %08x", urr->oid);
	  j->refetching = urr->oid;
	  if_modified_since = urr->access - clock_drift;
	}
      else if (now - urr->access > max_queue_time)
	{
	  log(L_INFO "%s: Max queueing time exceeded", j->url);
	  err = 601;
	  goto failed;
	}
    }
  else if (!j->fetching_robot_file)	/* Robots.txt not expected to be known for now */
    log(L_ERROR "Have to download URL %s which is not in the database", j->url);

  if (!okay_to_get(d, &u))
    {
      log(L_INFO "%s: filtered out", j->url);
      err = 600;
    failed:
      if (j->fetching_robot_file)
	{
	  set_attr_num(d, 'R', now);
	  goto restart0;
	}
      delete_object(j->refetching, j->url);
      urldb_store(j->url, now & ~1, err);
      dequeue_item(h);
      goto restart;
    }
  if (! (down = downloaders[u.protoid]))
    {
      log(L_INFO "%s: no downloader known", j->url);
      err = 602;
      goto failed;
    }

  j->qobj = qobj_new();			/* Create new object descriptor */

  rem_node(&j->n);			/* Fire! Run the job... */
  pid = fork();
  if (pid < 0)
    die("fork() failed: %m");
  if (!pid)
    {
      char buf[32];
      initlog(NULL);			/* Re-read PID */
      sprintf(buf, "%d", if_modified_since);
      if (trace_jobs)
	log(L_DEBUG "Running %s %s %d %d", down, j->url, if_modified_since, downgrade);
      if (trace_jobs > 1)
	log(L_INFO "<< %s (%s)", j->url, j->qobj->tf.name);
      close(1);				/* The object descriptor should be FD #1 */
      dup(j->qobj->tf.fh);
      close(j->qobj->tf.fh);
      execl(down, down, j->url, buf, downgrade ? "1" : NULL, NULL);
      die("exec %s failed: %m", down);
    }
  j->pid = pid;
  j->stat = JS_GATHER;
  add_tail(&running_jobs, &j->n);
  return;
}

static void
finished_bg_job(struct job *j, uns stat)
{
  int err;				/* Error status */
#define EST_OK 0			/* Got OK */
#define EST_TEMP 1			/* Temporary error */
#define EST_PERM 2			/* Permanent error */
#define EST_SAME -1			/* Not modified */
#define EST_ROBOT -2			/* Robot file */
  uns ecode;				/* Error code */
  FILE *f;
  byte *b, *c, *emsg;
  struct qitem *i = NULL;
  time_t tim = time(NULL);
  struct odes *d, *h;
  byte *url = j->url;

  if (trace_jobs)
    log(L_DEBUG "BG job %d exited with status %d", j->pid, stat);

  lseek(j->qobj->tf.fh, 0, SEEK_SET);	/* Load the object descriptor */
  f = fdopen(j->qobj->tf.fh, "r");
  if (!f)
    die("fdopen failed: %m");
  j->obj = obj_fload(f, j->qobj->tf.name);
  fclose(f);

  err = ecode = 0;			/* Are there any errors? */
  emsg = NULL;
  d = j->obj;
  h = j->hobj;
  set_attr(d, 'U', url);
  if (b = find_aval(d, '-'))		/* Fatal error? */
    {
      uns ec = (c = find_aval(h, 'E')) ? atol(c) : 0;
      set_attr_num(h, 'E', ec+1);
      err = EST_PERM;
      emsg = b;
      log(L_INFO "%s: [P] %s", url, emsg);
    }
  else if (b = find_aval(d, '?'))	/* Temp error? */
    {
      uns ec = (c = find_aval(h, 'e')) ? atol(c) : 0;
      set_attr_num(h, 'e', ec+1);
      ec = (c = find_aval(h, 'C')) ? atol(c) : 0;
      set_attr_num(h, 'C', ec+1);
      err = EST_TEMP;
      emsg = b;
      log(L_INFO "%s: [T] %s", url, emsg);
      if (b = find_aval(d, 'r'))	/* Retry-after present? */
	{
	  uns bb = atol(b);
	  set_attr_num(h, 'O', tim + bb);
	  log(L_INFO "%s: Will retry after %d seconds", url, bb);
	}
      if (b = find_aval(d, 'v'))	/* Downgrade? */
	{
	  set_attr_num(h, 'd', tim);
	  log(L_INFO "%s: Will downgrade protocol", url);
	}
      if (ec > max_rec_err)		/* Too many temp errors? */
	{
	  log(L_INFO "%s: [P] Too many recoverable errors, dropping", url);
	  err = EST_PERM;
	  ec = (c = find_aval(h, 'E')) ? atol(c) : 0;
	  set_attr_num(h, 'E', ec+1);
	}
    }
  else if ((b = find_aval(d, '0')) || find_aval(d, '+') || find_aval(d, 'X')) /* Downloaded OK or not modified */
    {
      uns ac = (c = find_aval(h, 'A')) ? atol(c) : 0;
      if (b)
	{
	  log(L_INFO "%s: Not modified", url);
	  err = EST_SAME;
	}
      else
	log(L_INFO "%s: OK", url);
      set_attr_num(h, 'A', ac+1);
      set_attr(h, 'C', NULL);
      set_attr(h, '-', NULL);
    }
  else					/* Missing status info */
    {
      log(L_ERROR "Incomplete object found");
      err = EST_TEMP;
      emsg = "0301 Incomplete object";
    }
  if (emsg)
    {
      sscanf(emsg, "%d", &ecode);
      set_attr(h, '-', emsg);
    }
  if (stat && !err)			/* Unknown error */
    {
      byte xxx[32];
      err = EST_TEMP;
      ecode = (stat < 1000) ? 9000 + stat : 9999;
      sprintf(xxx, "Error %d", ecode);
      set_attr(h, '-', xxx);
    }

  if (b = find_aval(d, 'S'))		/* Record server name */
    set_attr(h, 'S', b);

  set_attr_num(h, 'L', tim);		/* Update last access timestamp */

  if (err != EST_TEMP && j->fetching_robot_file) /* Got the access list file? */
    {
      parse_robot_file(h, err ? NULL : d);
      err = EST_ROBOT;			/* It was a robot file */
      set_attr_num(h, 'R', tim);
    }

  dump_host_obj(j->host->host_id, h);	/* Save the host descriptor and free the in-memory copy */
  obj_free(h);
  j->hobj = NULL;

  if (err != EST_ROBOT)			/* If not receiving a robot file... */
    i = dequeue_item(j->host);		/* ...dequeue the current item */
  if (err)				/* Errors or a robot file -> delete the files */
    {
      discard_objects(j);		/* Discard all incomplete objects and their data files */
      if (err == EST_TEMP)		/* Temp error => re-queue the item at the end */
	{
	  if (trace_jobs)
	    log(L_DEBUG "Re-queueing the item at the end");
#if 1
	  if (!i)
	    log(L_ERROR "BUG: EST_TEMP and no queued object for %s!", j->host);
	  else
#endif
	  enqueue_item(j->host, i->text);
	}
      else if (err == EST_PERM)		/* Mark the URL as permanently bad */
	{
	  urldb_store(url, tim & ~1, ecode);
	  delete_object(j->refetching, url);
	}
      else if (err == EST_SAME)		/* Mark the URL as updated */
	{
	  if (!j->refetching)
	    die("If-modified-since outside refetch process!");
	  urldb_store(url, tim & ~1, j->refetching);
	}
      else if (err == EST_ROBOT)	/* Robot file: Mark the URL as updated */
	urldb_store(url, tim & ~1, 8888);
      else
	die("Internal bug: invalid error flag %d", err);
    }
  else					/* OK -> move it to the object queue */
    {
      if (trace_jobs)
	log(L_DEBUG "Queueing the downloaded object for further processing");
      delete_object(j->refetching, url); /* Kill the old version if there's any */
      qobj_enqueue(j->qobj);
      j->qobj->obj = j->obj;
    }
  enqueue_host(j->host);

  rem_node(&j->n);
  j->stat = JS_FREE;
  add_tail(&free_jobs, &j->n);
}

static void
finished_job(pid_t pid, uns stat)
{
  if (fg_job && fg_job->pid == pid)
    finished_fg_job(stat);
  else
    {
      struct job *j;
      DO_FOR_ALL(j, running_jobs)
	if (j->pid == pid)
	  {
	    finished_bg_job(j, stat);
	    return;
	  }
      log(L_ERROR "Unknown process %d has exited with status %d", pid, stat);
    }
}

static void				/* Ignored... Only interrupts a system call. */
sig_alarm(int unused)
{
}

static void				/* Just sets the "shut_down" variable */
sig_shutdown(int unused)
{
  log(L_INFO "SHUTDOWN REQUESTED");
  shut_down = 1;
}

static void				/* The main loop */
loop(void)
{
  struct qhost *h;
  pid_t pid;
  int stat;
  uns wait_seconds;
  time_t now;
  struct sigaction siga;
  int done_fg_job;

  bzero(&siga, sizeof(siga));
  siga.sa_handler = sig_alarm;
  sigaction(SIGALRM, &siga, NULL);
  siga.sa_handler = sig_shutdown;
  siga.sa_flags = SA_RESTART;
  sigaction(SIGINT, &siga, NULL);
  sigaction(SIGTERM, &siga, NULL);
  sigaction(SIGHUP, &siga, NULL);
  for(;;)
    {
      now = time(NULL);
      wait_seconds = 0;
      while (!EMPTY_LIST(free_jobs) && qobj_count < max_queued_objs && !shut_down)
	{
	  h = dequeue_host(&wait_seconds, now);
	  if (!h)
	    break;
	  run_bg_job(h);
	}
      done_fg_job = 0;
      while (!fg_job && !shut_down && run_fg_job())
	done_fg_job = 1;
      if (!fg_job && EMPTY_LIST(running_jobs))
	{
	  if (shut_down)
	    break;
	  if (wait_seconds)
	    {
	      if (trace_jobs)
		log(L_DEBUG "Ready jobs must wait, sleeping for %d sec", wait_seconds);
	      sleep(wait_seconds);
	      continue;
	    }
	  else if (!done_fg_job)
	    {
	      log(L_INFO "All objects gathered.");
	      break;
	    }
	}
      if (wait_seconds)
	{
	  if (trace_jobs)
	    log(L_DEBUG "Ready jobs must wait, first wakeup after %d sec", wait_seconds);
	  alarm(wait_seconds);
	}
      pid = wait(&stat);
      if (pid < 0)
	{
	  if (errno == EINTR)
	    continue;
	  else
	    die("Wait: %m");
	}
      if (wait_seconds)
	alarm(0);
      if (WIFEXITED(stat))
	finished_job(pid, WEXITSTATUS(stat));
      else if (WIFSIGNALED(stat))
	{
	  log(L_ERROR "Process %d caught signal %d", pid, WTERMSIG(stat));
	  finished_job(pid, 900 + WTERMSIG(stat));
	}
      else
	{
	  log(L_ERROR "Process %d has died of unknown disease", pid);
	  finished_job(pid, 99999);
	}
    }
}

static void
load_refs(byte *name)
{
  FILE *f;
  byte buf[MAX_URL_SIZE];
  byte *b;

  log(L_INFO "Using %s as initial URL list", name);
  if (! (f = fopen(name, "r")))
    die("Cannot open initial URL list: %m");

  while (fgets(buf, MAX_URL_SIZE, f))
    {
      b = strchr(buf, '\n');
      if (!b)
	die("Initial URL list: line too long");
      *b = 0;
      if (!buf[0] || buf[0] == '#')
	continue;
      do_ref(buf, 1);
    }

  fclose(f);
}

int
main(int argc, char **argv)
{
  initlog(argv[0]);
  if (argc != 1 && argc != 2)
    die("Invalid arguments. Expecting: gatherd [<start-urls>]");
  
  read_config();
  db_init();
  url_filter_init();
  datatype_init();
  charset_init();

  log(L_INFO BANNER " starting...");
  gather_lock();
  queue_init();
  urldb_open();
  md5db_open();
  oqueue_init();
  job_init();

  if (argc == 2)
    load_refs(argv[1]);

  loop();

  log(L_INFO "Shutdown...");
  md5db_close();
  urldb_close();
  queue_cleanup();
  gather_unlock();
  return 0;
}
