/*
 *	Sherlock Database Builder -- Hashing of Non-Words
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>

#include <lib/lib.h>
#include <lib/searchdb.h>
#include <lib/fastbuf.h>

#undef NHASH_DEBUG

#ifdef NHASH_DEBUG
#define DBG(x, args...) log(L_DEBUG x, ## args)
#else
#define DBG(x, args...)
#endif

#ifdef SMALL
#define BUFSIZE 262144
#else
#define BUFSIZE 1048576
#endif

static struct fastbuf *fi, *fo, *idx;
static ulg count, size;
static uns tempno, nruns;

/* Configuration */

static uns magnification = 10;
static uns num_cache_words = 1024;
static uns cache_buf_size = 65536;

static struct cfitem cfitems[] = {
  { "Magnification", CI_INT, &magnification },
  { "CacheWords", CI_INT, &num_cache_words },
  { "CacheBuffer", CI_INT, &cache_buf_size },
  { NULL, CI_STOP, NULL }
};

/* Calculating hash table size... */

static void
sizing(void)
{
  bseek(fi, -sizeof(count), SEEK_END);
  count = bgetl_slow(fi);
  log(L_INFO "Building non-word hash table, %d items", count);
  size = nextprime(count * magnification / 10);
  log(L_INFO "Hash table size set to %d", size);
  bsetpos(fi, 0);
}

/* MERGESORT (shamelessly stolen from sword.c) */

struct fib {
  struct fastbuf *f;
  byte lastword[MAX_WORD_LEN];
  byte name[32];
  int eof;
  ulg hash, prevhash;
};

static struct fib files[4];
static struct fib *src[2], *dest[2];

static int pass_cnt = 0;

static void
opendest(void)
{
  int i;

  for(i=0; i<2; i++)
    {
      struct fib *x = dest[i];
      int fd;

      sprintf(x->name, "sdb/t%d/nhash-%d.%d-%d", tempno, getpid(), pass_cnt, i);
      fd = open(x->name, O_RDWR | O_CREAT | O_TRUNC, 0666);
      if (fd < 0)
	die("Unable to create %s: %m", x->name);
      x->f = bfdopen(fd, BUFSIZE);
      x->lastword[0] = x->eof = 0;
      DBG("Created output file `%s', channel #%d", x->name, i);
    }
  pass_cnt++;
}

static inline void
fill_in(struct fib *r)
{
  byte *p;
  int i;
  uns h = 0;
  struct fastbuf *f = r->f;

  if (r->eof || r->lastword[0])
    return;
  p = r->lastword;
  do
    {
      if (f->bptr < f->bstop)
	i = *f->bptr++;
      else if ((i = bgetc_slow(f)) == EOF)
	{
	  r->eof = 1;
	  r->lastword[0] = 0;
	  return;
	}
      *p++ = i;
      h = (h * HASH_MAGIC + i) % size;
    }
  while (i);
  r->hash = h;
}

static void
merge(void)
{
  struct fib *r1, *r2, *r, *w1, *w2;

  r1 = src[0];
  r2 = src[1];
  r1->lastword[0] = r2->lastword[0] = 0;
  if (!r2->f)
    r2->eof = 1;
  w1 = dest[0];
  w2 = dest[1];
  w1->hash = 0;
  nruns = 0;
  for(;;)
    {
      fill_in(r1);
      fill_in(r2);
      if (r1->eof && r2->eof)
	break;
      r1->prevhash = ~0;
      r2->prevhash = ~0;
      for(;;)
	{
	  int c1, c2;			/* Run continuation flags */

	  fill_in(r1);
	  fill_in(r2);
	  c1 = r1->eof ? 0 : (r1->prevhash == ~0U || r1->prevhash <= r1->hash);
	  c2 = r2->eof ? 0 : (r2->prevhash == ~0U || r2->prevhash <= r2->hash);
	  if (c1)			/* Select input channel... */
	    {
	      if (c2)
		{			/* Both channels non-empty */
		  if (r1->hash <= r2->hash)
		    r = r1;
		  else
		    r = r2;
		}
	      else
		r = r1;
	    }
	  else if (c2)
	    r = r2;
	  else				/* The world is coming to an end... Save your buffers... */
	    break;
	  if (w1->hash > r->hash)		/* Need to swap files */
	    {
	      struct fib *w;
	      w = w1; w1 = w2; w2 = w;
	      nruns++;
	    }
	  w1->hash = r->prevhash = r->hash;
	  w1->lastword[0] = 1;		/* MAGIC (mark output channel as used) */
	  bputs(w1->f, r->lastword);
	  bputc(w1->f, 0);
	  bputl(w1->f, bgetl(r->f));
	  r->lastword[0] = 0;
	}
    }
  bflush(w1->f);
  bflush(w2->f);
}

static void
closesrc(void)
{
  int i;

  for(i=0; i<2; i++)
    {
      struct fib *x = src[i];

      if (x->f)
	{
	  DBG("Closing `%s'", x->name);
	  bclose(x->f);
	  if (x->name[0] && unlink(x->name))
	    log(L_ERROR "unlink(`%s'): %m", x->name);
	}
    }
}

static void
swapsrcdest(void)
{
  struct fib *a;

  DBG("Swapping...");
  a = src[0]; src[0] = dest[0]; dest[0] = a;
  a = src[1]; src[1] = dest[1]; dest[1] = a;
  bsetpos(src[0]->f, 0);
  bsetpos(src[1]->f, 0);
}

static void
closefinal(void)
{
  struct fib *x;
  ulg lasth = ~0;
  ulg ref;
  uns coll = 0;
  uns total = 0;
  uns maxcoll = 0;

  swapsrcdest();
  DBG("Final output...");
  x = src[0];
  x->lastword[0] = 0;
  for(;;)
    {
      fill_in(x);
      if (x->eof)
	break;
      ref = bgetl(x->f);
      if (x->hash != lasth)
	{
	  if (lasth != ~0U)
	    bputc(fo, 0);
	  while (++lasth < x->hash)
	    bputl(idx, ~0);
	  bputl(idx, btell(fo));
	  total++;
	  coll = 0;
	}
      else if (++coll > maxcoll)
	maxcoll = coll;
      bputs(fo, x->lastword);
      bputc(fo, 0);
      bputl(fo, ref);
      x->lastword[0] = 0;
    }
  bputc(fo, 0);
  while (++lasth < size)
    bputl(idx, ~0);
  log(L_INFO "Hash table built. Used %d slots, got maximal chain of %d items.", total, maxcoll+1);
  closesrc();				/* Delete everything */
}

/* Pre-sorting cache */

struct cache_item {
  ulg ref;
  ulg hash;
  byte word[1];
};

static byte *cache_buffer, *cache_buf_end, *cache_free;
static struct cache_item **cache_index;

static int
ci_cmp(struct cache_item **a, struct cache_item **b)
{
  struct cache_item *A = *a;
  struct cache_item *B = *b;

  if (A->hash < B->hash)
    return -1;
  if (A->hash > B->hash)
    return 1;
  return 0;
}

static void
presort(void)
{
  struct fib *r, *w;
  uns wi, wcnt, l;
  struct cache_item *i;

  cache_free = cache_buffer = xmalloc(cache_buf_size);
  cache_buf_end = cache_buffer + cache_buf_size;
  cache_index = xmalloc(num_cache_words * sizeof(struct cache_item *));
  wcnt = wi = 0;

  r = src[0];
  nruns = 0;
  while (!r->eof)
    {
      w = dest[wi];

      for(;;)
	{
	  fill_in(r);
	  if (r->eof)
	    break;
	  if (wcnt >= num_cache_words)
	    break;
	  l = sizeof(struct cache_item) + strlen(r->lastword);
	  l = (l + CPU_STRUCT_ALIGN - 1) & ~(CPU_STRUCT_ALIGN - 1);
	  if (cache_free + l > cache_buf_end)
	    break;
	  i = (struct cache_item *) cache_free;
	  cache_free += l;
	  i->ref = bgetl(r->f);
	  i->hash = r->hash;
	  strcpy(i->word, r->lastword);
	  cache_index[wcnt++] = i;
	  r->lastword[0] = 0;
	}

      if (wcnt > 0)
	{
	  qsort(cache_index, wcnt, sizeof(struct cache_item *), (int (*)(const void *, const void *)) ci_cmp);
	  for(l=0; l<wcnt; l++)
	    {
	      i = cache_index[l];
	      bputs(w->f, i->word);
	      bputc(w->f, 0);
	      bputl(w->f, i->ref);
	    }
	  cache_free = cache_buffer;
	  wcnt = 0;
	}

      w->lastword[0] = 1;		/* MAGIC */
      wi = !wi;
      nruns++;
    }

  bflush(dest[0]->f);
  bflush(dest[1]->f);
  free(cache_buffer);
  free(cache_index);
}

int
main(int argc, char **argv)
{
  int pass = 0;

  initlog(argv[0]);
  if (argc != 4)
    die("Usage: nhash <nonword-list> <hash-list> <hash-index>");
  fi = bopen(argv[1], O_RDONLY, BUFSIZE);
  fo = bopen(argv[2], O_WRONLY | O_CREAT | O_TRUNC, BUFSIZE);
  idx = bopen(argv[3], O_WRONLY | O_CREAT | O_TRUNC, BUFSIZE);
  cf_read("cf/nhash", cfitems);
  sizing();
  src[0] = files;
  src[1] = files + 1;
  dest[0] = files + 2;
  dest[1] = files + 3;
  src[0]->f = fi;
  for(;;)
    {
      opendest();
      if (pass++)
	merge();
      else
	presort();
      closesrc();
      if (!dest[1]->lastword[0])
	break;
      swapsrcdest();
      log(L_DEBUG "Pass %d: %d runs.", pass_cnt, nruns);
      tempno = !tempno;
    }
  log(L_DEBUG "Closing files.");
  closefinal();
  bclose(fo);
  bclose(idx);
  log(L_DEBUG "Done in %d passes.", pass_cnt);
  return 0;
}
