/*
 *	Sherlock Database Builder -- Word Reference File Sorting
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include <lib/string.h>
#include <lib/lib.h>
#include <lib/searchdb.h>
#include <lib/wordcmp.h>
#include <lib/fastbuf.h>

#undef SWORD_DEBUG

#ifdef SWORD_DEBUG
#define DBG(x, args...) log(L_DEBUG x, ## args)
#else
#define DBG(x, args...)
#endif

#ifdef SMALL
#define BUFSIZE 262144
#else
#define BUFSIZE 1048576
#endif

struct fib {
  struct fastbuf *f;
  byte lastword[MAX_WORD_LEN];
  byte prevword[MAX_WORD_LEN];
  byte name[32];
  int eof;
};

static struct fib files[4];
static struct fib *src[2], *dest[2];
static uns nonword_flag, delete_flag;
static uns tempno, nruns;

static inline int
compare(byte *a, byte *b)
{
  return nonword_flag ? strcmp(a, b) : search_wordcmp(a, b);
}

static int pass_cnt = 0;

static void
opendest(void)
{
  int i;

  for(i=0; i<2; i++)
    {
      struct fib *x = dest[i];
      int fd;

      sprintf(x->name, "sdb/t%d/sword-%d.%d-%d", tempno, getpid(), pass_cnt, i);
      fd = open(x->name, O_RDWR | O_CREAT | O_TRUNC, 0666);
      if (fd < 0)
	die("Unable to create %s: %m", x->name);
      x->f = bfdopen(fd, BUFSIZE);
      x->lastword[0] = x->eof = 0;
      DBG("Created output file `%s', channel #%d", x->name, i);
    }
  pass_cnt++;
}

static inline void
fill_in(struct fib *r)
{
  byte *p;
  int i;
  struct fastbuf *f = r->f;

  if (r->eof || r->lastword[0])
    return;
  p = r->lastword;
  do
    {
      if (f->bptr < f->bstop)
	i = *f->bptr++;
      else if ((i = bgetc_slow(f)) == EOF)
	{
	  r->eof = 1;
	  r->lastword[0] = 0;
	  return;
	}
      *p++ = i;
    }
  while (i);
}

static void
merge(void)
{
  struct fib *r1, *r2, *r, *w1, *w2;
  int flag;

  nruns = 1;
  r1 = src[0];
  r2 = src[1];
  r1->lastword[0] = r2->lastword[0] = 0;
  if (!r2->f)
    r2->eof = 1;
  w1 = dest[0];
  w2 = dest[1];
  for(;;)
    {
      fill_in(r1);
      fill_in(r2);
      if (r1->eof && r2->eof)
	break;
      r1->prevword[0] = 0;
      r2->prevword[0] = 0;
      for(;;)
	{
	  int c1, c2;			/* Run continuation flags */

	  fill_in(r1);
	  fill_in(r2);
	  c1 = r1->eof ? 0 : (!r1->prevword[0] || compare(r1->prevword, r1->lastword) < 0);
	  c2 = r2->eof ? 0 : (!r2->prevword[0] || compare(r2->prevword, r2->lastword) < 0);
	  flag = 0;
	  if (c1)			/* Select input channel... */
	    {
	      if (c2)
		{			/* Both runs continue */
		  int c = compare(r1->lastword, r2->lastword);
		  if (c < 0)
		    r = r1;
		  else if (c > 0)
		    r = r2;
		  else			/* Need to merge! */
		    {
		      r = r1;
		      flag = 1;
		    }
		}
	      else
		r = r1;
	    }
	  else if (c2)
	    r = r2;
	  else				/* The world is coming to an end... Save your buffers... */
	    break;
	  if (w1->lastword[0] && compare(w1->lastword, r->lastword) >= 0) /* Need to swap files */
	    {
	      struct fib *w;
	      w = w1; w1 = w2; w2 = w;
	      nruns++;
	    }
	  strcpy(w1->lastword, r->lastword);
	  bputs(w1->f, r->lastword);
	  bputc(w1->f, 0);
	  if (!flag)			/* Copy only */
	    {
	      ulg l;
	      l = bgetl(r->f);
	      bputl(w1->f, l);
	      bbcopy(r->f, w1->f, l * sizeof(struct sdbitem));
	      strcpy(r->prevword, r->lastword);
	      r->lastword[0] = 0;
	    }
	  else				/* Merge */
	    {
	      ulg l1, l2, l;
	      struct sdbitem i1, i2;
	      l1 = bgetl(r1->f);
	      l2 = bgetl(r2->f);
	      if (!l1 || !l2)
		die("Internal error: zero-length reference list spotted!");
	      l = l1 + l2;
	      bputl(w1->f, l);
	      bread(r1->f, &i1, sizeof(i1));
	      bread(r2->f, &i2, sizeof(i2));
	      while (l1 && l2)
		{
		  if (i1.oid < i2.oid
		      || i1.oid == i2.oid && i1.flags < i2.flags)
		    {			/* If OID's match, tags must match, too (!!) */
		      bwrite(w1->f, &i1, sizeof(i1));
		      if (--l1)
			bread(r1->f, &i1, sizeof(i1));
		    }
		  else
		    {
		      bwrite(w1->f, &i2, sizeof(i2));
		      if (--l2)
			bread(r2->f, &i2, sizeof(i2));
		    }
		}
	      if (l1)
		{
		  bwrite(w1->f, &i1, sizeof(i1));
		  bbcopy(r1->f, w1->f, (l1-1) * sizeof(struct sdbitem));
		}
	      else if (l2)
		{
		  bwrite(w1->f, &i2, sizeof(i2));
		  bbcopy(r2->f, w1->f, (l2-1) * sizeof(struct sdbitem));
		}
	      strcpy(r1->prevword, r1->lastword);
	      strcpy(r2->prevword, r2->lastword);
	      r1->lastword[0] = r2->lastword[0] = 0;
	    }
	}
    }
  bflush(w1->f);
  bflush(w2->f);
}

static void
closesrc(void)
{
  int i;

  for(i=0; i<2; i++)
    {
      struct fib *x = src[i];

      if (x->f)
	{
	  DBG("Closing `%s'", x->name);
	  bclose(x->f);
	  if (x->name[0] && unlink(x->name))
	    log(L_ERROR "unlink(`%s'): %m", x->name);
	}
    }
}

static void
swapsrcdest(void)
{
  struct fib *a;

  DBG("Swapping...");
  a = src[0]; src[0] = dest[0]; dest[0] = a;
  a = src[1]; src[1] = dest[1]; dest[1] = a;
  bsetpos(src[0]->f, 0);
  bsetpos(src[1]->f, 0);
}

static void
closefinal(byte *name)
{
  struct fib *x = dest[0];

  bclose(x->f);
  DBG("Renaming `%s' to dest file `%s'", x->name, name);
  if (rename(x->name, name))
    {
      int f0, f1, l;
      byte *buf = xmalloc(BUFSIZE);

      DBG("Doing manual copy of `%s' to `%s'", x->name, name);
      f0 = open(x->name, O_RDONLY);
      if (f0 < 0)
	die("open(%s): %m", x->name);
      f1 = open(name, O_WRONLY | O_CREAT | O_TRUNC, 0666);
      if (f1 < 0)
	die("open(%s): %m", x->name);
      while ((l = read(f0, buf, BUFSIZE)) > 0)
	if (write(f1, buf, l) != l)
	  die("write: %m");
      if (l < 0)
	die("read: %m");
      close(f0);
      close(f1);
      if (unlink(x->name))
	die("unlink(`%s'): %m", x->name);
    }

  x = dest[1];
  DBG("And unlinking dest `%s'", x->name);
  bclose(x->f);
  if (unlink(x->name))
    die("unlink(`%s'): %m", x->name);
}

int
main(int argc, char **argv)
{
  initlog(argv[0]);
  if (argc > 1 && !strcmp(argv[1], "-n"))
    {
      nonword_flag = 1;
      argc--;
      argv++;
    }
  if (argc > 1 && !strcmp(argv[1], "-d"))
    {
      delete_flag = 1;
      argc--;
      argv++;
    }
  if (argc < 3)
    die("Usage: sword [-n] [-d] <src> <dest>");
  log(L_INFO "Sorting %sword references", nonword_flag ? "non-" : "");
  src[0] = files;
  src[1] = files + 1;
  dest[0] = files + 2;
  dest[1] = files + 3;
  src[0]->f = bopen(argv[1], O_RDONLY, BUFSIZE);
  if (delete_flag)
    strcpy(src[0]->name, argv[1]);
  for(;;)
    {
      opendest();
      merge();
      closesrc();
      if (!dest[1]->lastword[0])
	break;
      swapsrcdest();
      log(L_DEBUG "Pass %d: %d runs.", pass_cnt, nruns);
      tempno = !tempno;
    }
  log(L_DEBUG "Closing files.");
  closefinal(argv[2]);
  log(L_DEBUG "Done in %d passes.", pass_cnt);
  return 0;
}
