/*
 *	Sherlock Search Engine -- File Page Cache
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>

#undef BIG_DEBUG
#undef DETAIL_DEBUG

#include "sherlock.h"

#include <lib/searchdb.h>

struct file_page {
  byte data[FILE_PAGE_SIZE];
};

struct file_page_info {
  node h;				/* Hash chain (_must_ be first!) */
  node n;				/* LRU chain */
  ulg key;				/* Position + filedes */
  uns lockcnt;				/* Lock counter */
};

static struct file_page *fc_pages;
static struct file_page_info *page_infos;
static list fc_lru, fc_free, *fc_hash;
static uns fc_hash_count;
static struct file_page_info *last_page;

#ifdef DETAIL_DEBUG
static uns lock_count;
#endif

void
file_cache_init(void)
{
  uns i;

  fc_hash_count = nextprime(file_cache_size / 8);
  log(L_INFO "Allocating file cache for %d entries hashed to %d boxes", file_cache_size, fc_hash_count);
  fc_pages = xmalloc(sizeof(struct file_page) * file_cache_size);
  init_list(&fc_lru);
  init_list(&fc_free);
  page_infos = xmalloc(sizeof(struct file_page_info) * file_cache_size);
  for(i=0; i<file_cache_size; i++)
    add_tail(&fc_free, &page_infos[i].h);
  fc_hash = xmalloc(sizeof(list) * fc_hash_count);
  for(i=0; i<fc_hash_count; i++)
    init_list(fc_hash+i);
}

int
open_file(byte *name)
{
  int x = open(name, O_RDONLY);
  if (x < 0)
    die("Unable to open `%s': %m", name);
  DBG("File %d is %s", x, name);
  return x;
}

byte *
read_page(int fd, ulg pos)
{
  ulg key = pos + fd;
  uns hash = key % fc_hash_count;
  struct file_page_info *i;
  byte *p;

  if (pos & (FILE_PAGE_SIZE-1))
    log(L_ERROR "Misaligned read_page(%d,%x)!", fd, pos);
  DO_FOR_ALL(i, fc_hash[hash])
    if (i->key == key)
      {
	DDBG("%d/%d: found in cache [lock=%d]", fd, pos, i->lockcnt);
	if (!i->lockcnt)
	  {
	    rem_node(&i->n);
	    add_tail(&fc_lru, &i->n);
	  }
	last_page = i;
	return (byte *)(fc_pages + (i - page_infos));
      }

  i = HEAD(fc_free);
  if (!i->h.next)
    {
      i = SKIP_BACK(struct file_page_info, n, HEAD(fc_lru)); /* :-( */
      if (!i->n.next)
	die("read_page() out of slots");
      rem_node(&i->n);
      DDBG("Reusing page %08x for %08x, lock=%d", i->key, key, i->lockcnt);
      if (i->lockcnt)
        log(L_ERROR "Reusing locked page!!!");
    }
  i->lockcnt = 0;
  DDBG("%d/%d: reading [%08x]", fd, pos, key);
  rem_node(&i->h);
  add_tail(&fc_hash[hash], &i->h);
  add_tail(&fc_lru, &i->n);
  i->key = key;
  p = (byte *)(fc_pages + (i - page_infos));
  if (lseek(fd, pos, SEEK_SET) < 0)
    die("Seek to %d failed on file %d: %m", pos, fd);
  if (read(fd, p, FILE_PAGE_SIZE) <= 0)
    die("Read on file %d failed: %m", fd);
  last_page = i;
  return p;
}

byte *
read_data(int fd, ulg pos, uns *len)
{
  ulg pos0 = pos & ~(FILE_PAGE_SIZE - 1);
  ulg pos1 = pos & (FILE_PAGE_SIZE - 1);
  *len = FILE_PAGE_SIZE - pos1;
  return read_page(fd, pos0) + pos1;
}

void *
lock_last_page(void)
{
  struct file_page_info *i = last_page;

  DDBG("Locking page %x (cnt=%d,locked=%d)", i->key, i->lockcnt+1, ++lock_count);
  if (!i->lockcnt++)
    rem_node(&i->n);			/* Remove from the LRU queue */
  return i;
}

void
unlock_page(void *x)
{
  struct file_page_info *i = x;

  DDBG("Unlocking page %x (cnt=%d,locked=%d)", i->key, i->lockcnt-1, --lock_count);
  if (!--i->lockcnt)
    add_tail(&fc_lru, &i->n);
}
