/*
 *	Sherlock Search Engine -- Searching for Words
 *
 *	(c) 1997, 1998 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

/*
 *	Think twice before touching. And then don't.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <alloca.h>

#include "sherlock.h"

#include <lib/searchdb.h>
#include <lib/string.h>
#include <lib/wordcmp.h>
#include <lib/pools.h>

int have_word_list;

static int tree_fd, wlist_fd;
static struct tree_page tree_root;

/* BTREE */

#define MAX_DEPTH 4

static uns root_pointers[KEYS_PER_PAGE + 1], root_sizes[KEYS_PER_PAGE + 1],
           rel_pointers[KEYS_PER_PAGE + 1][MAX_DEPTH], depths[KEYS_PER_PAGE + 1];

static void
btree_init(void)
{
  struct tree_page *root;
  uns keys, skip, i, j, w, k, p, ww, pos, rw, depth;
  uns width[MAX_DEPTH];

  root = (struct tree_page *) read_data(tree_fd, 0, &i);
  memcpy(&tree_root, root, sizeof(tree_root));
  keys = root->aux;			/* See search/build/wtree.c for an explanation of what we're doing here */
  skip = keys / (KEYS_PER_PAGE + 1);
  i = 0;
  pos = TREE_PAGE_SIZE;
  for(j=0; j<=KEYS_PER_PAGE; j++)
    {
      w = skip;
      if (i >= keys)
	break;
      if (i + w >= keys)
	w = keys - i;			/* Subtree: <w> keys starting with the <i>-th */

      width[0] = 0;			/* Calculate subtree width */
      depth = 0;
      while (width[depth] < w)
	{
	  width[depth+1] = width[depth] * (KEYS_PER_PAGE+1) + KEYS_PER_PAGE;
	  depth++;
	}
      depths[j] = depth;
      if (depth > MAX_DEPTH)
	die("Tree too high to climb on");
      p = 0;				/* Total pages per subtree */
      rw = 1;
      for(k=0; k<depth; k++)
	{
	  rel_pointers[j][k] = p;
	  ww = width[depth - k - 1] * (KEYS_PER_PAGE + 1) + KEYS_PER_PAGE + 1;
	  if (k == depth - 1)
	    p += (w + ww - 2) / ww;
	  else
	    p += rw;
	  rw *= (KEYS_PER_PAGE + 1);
	}
      root_sizes[j] = p;
      root_pointers[j] = pos;
      DDBG("subtree=%d first=%d keys=%d depth=%d pages=%d pkeys=%d pos=%d len=%d",
	  j, i, w, depth, p, p*KEYS_PER_PAGE, pos, p * TREE_PAGE_SIZE);
      pos += p * TREE_PAGE_SIZE;
      i += skip + 1;
    }
  DBG("Total tree file size should be %d", pos);
  if ((uns) lseek(tree_fd, 0, SEEK_END) != pos)
    die("Unexpected inconsistency in the btree: length should be %d, but it isn't!", pos);
}

static int
lookup_bt_key(struct tree_page *p, byte *key)
{
  int l = 0;
  int r = KEYS_PER_PAGE - 1;
  int m = 0;
  int z;

  while (l <= r)
    {
      m = (l + r) / 2;
      DDBG("/%-8.8s/%-8.8s/", p->keys[m].sig, key);
      z = memcmp(key, p->keys[m].sig, 8);
      DDBG("l=%d m=%d r=%d z=%d", l,m,r,z);
      if (!z)
	return -m-1;
      if (z < 0)
	r = m - 1;
      else
	l = m + 1;
    }
  z = memcmp(key, p->keys[m].sig, 8);
  return (z > 0) ? (m + 1) : m;
}

static ulg
find_sig(byte *sig)
{
  struct tree_page *r = &tree_root;
  int i, z;
  uns base, pag, npag, dep;

  i = z = lookup_bt_key(r, sig);
  if (i < 0)
    goto found;
  base = root_pointers[z];
  npag = root_sizes[z];
  pag = 0;
  dep = 0;
  DDBG("Going to subtree %d, base=%d, npag=%d", z, base, npag);
  while (pag < npag)
    {
      r = (struct tree_page *) read_data(tree_fd, TREE_PAGE_SIZE*pag + base, &i);
      i = lookup_bt_key(r, sig);
      if (i < 0)
	goto found;
      dep++;
      if (dep > depths[z])
	{
	  DDBG("Too deep. [DG]iving up.");
	  return ~0;
	}
#if 0
      pag = rel_pointers[z][dep] + pag * (KEYS_PER_PAGE+1) + i;
#else
      pag = pag * (KEYS_PER_PAGE+1) + i + 1;
#endif
      DDBG("Going to page %d", pag);
    }
  DDBG("Fell off the tree");
  return ~0;

found:
  i = -i-1;
  DDBG("Found at position %d, sigpos=%d", i, r->keys[i].ptr);
  return r->keys[i].ptr;
}

/* Filtering of words */

static inline int
match_detail(byte *buf, word *uni, word *uni_unacc, uns wlen)
{
  word uc[MAX_WORD_LEN], uc_unacc[MAX_WORD_LEN];
  word *uu, *uua;
  int accents = 0;

  uu = uc;
  uua = uc_unacc;
  while (*buf)
    {
      uns u, ua;
      GET_UTF8(buf, u);
      ua = Uunaccent(u);
      *uu++ = u;
      *uua++ = ua;
      if (u != ua)
	accents++;
    }
  if (uu - uc != (int) wlen)		/* Lengths differ -> shortcut */
    return 0;
  if (uni_unacc && accents)		/* Both accented -> compare accented forms */
    return !memcmp(uni, uc, sizeof(word) * wlen);
  else					/* Any of them without accents -> compare unaccented forms */
    return !memcmp(uni_unacc ? : uni, uc_unacc, sizeof(word) * wlen);
}

static int
lookup_detail(word *uni, word *uni_unacc, uns len, ulg *x, ulg pos)
{
  uns cnt = 0;
  byte buf[MAX_WORD_LEN];
  byte *b = buf;
  byte *r, *e;
  uns s = 0;
  uns l, wl = 0;
  ulg ref;

  for(;;)
    {
      r = read_data(wlist_fd, pos, &l);
      pos += l;
      e = r + l;
      while (r < e)
	{
	  if (s)
	    {
	      *b++ = *r++;
	      if (!--s)
		{
		  DDBG("/%s/ %d", buf, ref);
		  if (match_detail(buf, uni, uni_unacc, len))
		    {
		      if (cnt++ >= max_word_matches)
			{
			  DBG("Too many matches");
			  return -1;
			}
		      *x++ = ref;
		      DDBG("MATCH");
		    }
		  b = buf;
		}
	    }
	  else if (*r)
	    *b++ = *r++;
	  else
	    {
	      *b = *r++;
	      wl = b - buf;
	      if (!wl)
		return cnt;
	      b = (byte *) &ref;
	      s = sizeof(ref);
	    }
	}
    }
}

/* Main word lookup function */

int
lookup_word(byte *w, ulg *x)
{
  int accents = 0;
  byte signature[8];
  ulg sigpos;
  uns wlen;
  word *uni, *uni_unacc, *uu, *uua;

  wlen = strlen(w) + 1;
  uu = uni = alloca(wlen * sizeof(word));
  uua = uni_unacc = alloca(wlen * sizeof(word));
  while (*w)
    if (*w == '?' || *w == '*')
      {
	log(L_ERROR "Wildcard search requested, but not supported.");
	return 0;
      }
    else if (Cspace(*w))
      {
	DBG("Shortcut: spaces -> not a word");
	return 0;
      }
    else
      {
	uns uc, ua;
	GET_UTF8(w, uc);
	ua = Uunaccent(uc);
	*uu++ = uc;
	*uua++ = ua;
	if (uc != ua)
	  accents++;
      }
  wlen = uu - uni;
  *uu = *uua = 0;

  extract_sig_uni(uni, signature);
  sigpos = find_sig(signature);
  if (sigpos == ~0U)
    return 0;

  if (accents)
    DBG("Unaccented: `%s'", static_ucs2_to_utf8(uni_unacc));
  else
    uni_unacc = NULL;

  return lookup_detail(uni, uni_unacc, wlen, x, sigpos);
}

void
word_init(void)
{
  if (sig_tree_name || word_list_name)
    {
      have_word_list = 1;
      tree_fd = open_file(sig_tree_name);
      wlist_fd = open_file(word_list_name);
      btree_init();
    }
}

byte *
normalize_word(struct mempool *p, byte *x)
{
  byte buf[MAX_WORD_LEN];
  byte *d = buf;
  uns len;

  while (*x)
    {
      uns uc;
      GET_UTF8(x, uc);
      uc = Utoupper(uc);
      PUT_UTF8(d, uc);
      if (d > buf + MAX_WORD_LEN - 3)
	break;
    }
  len = d - buf;
  x = pool_alloc(p, len+1);
  memcpy(x, buf, len);
  x[len] = 0;
  return x;
}
